""" POO - Python-Only ORB (Object Request Broker)
Copyright (C) 2002 James ''Wez''Weatherall

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""

import threading
import socket
import pickle
import sys
import time
import string

class _sfile:
    def __init__(self, sock):
        self.s=sock
        self.b=""
        self.ob=[]
    def read(self, l):
        self.flush()
        if len(self.b) >= l:
            d=self.b[:l]
            self.b=self.b[l:]
        else:
            d=self.b
            while len(d) < l:
                data=self.s.recv(1024)
                if not data: break
                n=l-len(d)
                d=d+data[:n]
                self.b=data[n:]
        return d
    def readline(self):
        self.flush()
        while "\n" not in self.b:
            data=self.s.recv(1024)
            if not data: return ""
            self.b=self.b+data
        i=string.find(self.b, "\n")+1
        d=self.b[:i]
        self.b=self.b[i:]
        return d
    def write(self, data):
        self.ob.append(data)
        if "\n" in data: self.flush()
    def flush(self):
        d=string.join(self.ob, "")
        if (d):
            self.s.send(d)
            self.ob=[]
    def __del__(self):
        self.flush()

class ORB:
    class _dispatcher(threading.Thread):
        def __init__(self, sock, orb):
            self.sock = sock
            self.orb = orb
            orb.dispatchers.append(self)
            threading.Thread.__init__(self)
            self.setDaemon(1)
        def run(self):
            sock = self.sock
            orb = self.orb
            self.active=1
            try:
                while self.active:
                    conn, client = sock.accept()
                    worker = orb.new_worker(conn)
            finally:
                orb.dispatchers.remove(self)
                self.orb=None
        def kill(self):
            self.active=0
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect(self.sock.getsockname())
            except:
                pass
            self.sock.close()
            self.join()
            
    class _worker(threading.Thread):
        def __init__(self, sock, orb):
            sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
            self.sock = sock
            self.orb = orb
            orb.workers.append(self)
            threading.Thread.__init__(self)
            self.setDaemon(1)
        def run(self):
            file = _sfile(self.sock)
            orb = self.orb
            mtype = type(self.run)
            self.active=1
            try:
                while self.active:
                    try:
                        key, method, params = pickle.load(file)
                    except:
                        print "worker error!", repr(sys.exc_info()[1])
                        break
                    try:
                        obj=orb.servants[key]
                    except:
                        print "no such object"
                    try:
                        if method == "\000get":
                            value=getattr(obj, params[0])
                            if type(value) == mtype:
                                pickle.dump((0, (1, None)), file)
                            else:
                                pickle.dump((0, (0, value)), file)
                        elif method == "\000set":
                            pickle.dump((0, apply(setattr, (obj,)+params)), file)
                        else:
                            value=getattr(obj, method)
                            pickle.dump((0, apply(value, params)), file)
                    except:
                        pickle.dump((1, sys.exc_info()[1]), file)
            finally:
                orb.workers.remove(self)
                self.orb=None
        def kill(self):
            self.active=0
            self.sock.close()

    def __init__(self):
        self.servants={}
        self.newkey=1
        self.active=1
        self.dispatchers=[]
        self.workers=[]
    def add_servant(self, object):
        if object not in self.servants.values():
            key=self.newkey
            self.newkey=self.newkey+1
            self.servants[key] = object
            object._poo_key = key
            object._poo = self
            object._poo_stub = stub(map(lambda x:x.sock.getsockname(), self.dispatchers), key)
	return object._poo_stub
    def rem_servant(self, object):
        if object in self.servants.values():
            del self.servants[object._poo_key]
    def shutdown(self):
        self.active=0
        for thread in self.dispatchers:
            thread.kill()
        for thread in self.workers:
            thread.kill()
        self.servants={}

    def new_dispatcher(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(("", 0))
        sock.listen(5)
        thread=self._dispatcher(sock, self)
        thread.start()
    def new_worker(self, sock):
        thread=self._worker(sock, self)
        thread.start()

class stub:
    class _fn:
        def __init__(self, stub, name):
            self.stub=stub
            self.name=name
        def __call__(self, *args):
            file=self.stub._stub__sfile()
            pickle.dump((self.stub._stub__key, self.name, args), file)
            result=pickle.load(file)
            self.stub._stub__socks.append(file)
            if result[0] == 0:
                return result[1]
            else:
                raise result[1]
    def __init__(self, addresses, key):
        self.__stubattr("addrs", addresses)
        self.__stubattr("key", key)
        self.__stubattr("socks", [])
    def __stubattr(self, name, value):
        self.__dict__["_stub__"+name]=value
    def __sfile(self):
        try:
            file=self.__socks.pop()
        except:
            sock=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.bind(("", 0))
            sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
            sock.connect(self.__addrs[0])
            file=_sfile(sock)
        return file
    def __rpc(self, name, params):
        file=self.__sfile()
        pickle.dump((self.__key, name, params), file)
        result=pickle.load(file)
        self.__socks.append(file)
        if result[0] == 0:
            return result[1]
        else:
            raise result[1]        
    def __getstate__(self):
        return {"addrs":self.__addrs,
		"key":self.__key}
    def __setstate__(self, state):
        self.__stubattr("socks", [])
        self.__stubattr("addrs", state["addrs"])
        self.__stubattr("key", state["key"])
    def __getattr__(self, name):
        kind, value = self.__rpc("\000get", (name,))
        if kind == 0:
            return value
        elif kind == 1:
            return self._fn(self, name)
    def __setattr__(self, name, value):
        return self.__rpc("\000set", (name, value))
    def __hasattr__(self, name):
        return self.__rpc("\000has", (name,))

if __name__=="__main__":
    print "Testing..."
    orb=ORB()
    try:
        orb.new_dispatcher()

        class test_object:
            def __len__(self):
                return 0
            def __call__(self):
                return "I contain \"%s\"" % self.data
            def echo(self, string):
                self.data=string
                return string
            data=""

        print "Creating object"
        stub = orb.add_servant(test_object())

        print "Calling object"
        t1=time.time()
        N=1000
        for i in range(N):
            value=stub.echo("Hello World")
        t2=time.time()
        print "data=", stub.data
        stub.data="This is a test"
        print "data=", stub.data
        print "calling...", stub()
        print "length=", len(stub)
        print "Time:", t2-t1, " - ", (t2-t1)/N, "per invocation"
    finally:
        print "shutdown"
	del stub
        orb.shutdown()
        print "done"

